# region HEADER
# ---------------------------------------------------------------------------------------
#
#   raygui v2.0 - A simple and easy-to-use immedite-mode-gui library for raylib (www.raylib.com)
#
#   DESCRIPTION:
#
#   raygui is a library for creating simple IMGUI interfaces using raylib, intended for tools dev. 
#   
#   Basic controls provided:
#
#       - WindowBox
#       - GroupBox
#       - Line
#       - Panel
#       - Label
#       - Button
#       - LabelButton
#       - ImageButton
#       - ImageButtonEx
#       - ToggleButton
#       - ToggleGroup
#       - CheckBox
#       - ComboBox
#       - DropdownBox
#       - Spinner
#       - ValueBox
#       - TextBox
#       - TextBoxMulti
#       - Slider
#       - SliderBar
#       - ProgressBar
#       - StatusBar
#       - ScrollPanel
#       - ListView
#       - ColorPicker
#       - MessageBox
#       - DummyRec
#       - Grid
#
#   It also provides a set of functions for styling the controls based on its properties (size, color).
#
#   CONFIGURATION:
#
#   #define RAYGUI_IMPLEMENTATION
#       Generates the implementation of the library into the included file.
#       If not defined, the library is in header only mode and can be included in other headers
#       or source files without problems. But only ONE file should hold the implementation.
#
#   #define RAYGUI_STATIC (defined by default)
#       The generated implementation will stay private inside implementation file and all
#       internal symbols and functions will only be visible inside that file.
#
#   #define RAYGUI_STANDALONE
#       Avoid raylib.h header inclusion in this file. Data types defined on raylib are defined
#       internally in the library and input management and drawing functions must be provided by
#       the user (check library implementation for further details).
#
#   #define RAYGUI_STYLE_SAVE_LOAD
#       Include style customization and save/load functions, useful when required.
#
#
#   VERSIONS HISTORY:
#       2.0 (01-May-2018) Lot of rework and redesign! Lots of new controls!
#       1.5 (21-Jun-2017) Working in an improved styles system
#       1.4 (15-Jun-2017) Rewritten all GUI functions (removed useless ones)
#       1.3 (12-Jun-2017) Redesigned styles system
#       1.1 (01-Jun-2017) Complete review of the library
#       1.0 (07-Jun-2016) Converted to header-only by Ramon Santamaria.
#       0.9 (07-Mar-2016) Reviewed and tested by Albert Martos, Ian Eito, Sergio Martinez and Ramon Santamaria.
#       0.8 (27-Aug-2015) Initial release. Implemented by Kevin Gato, Daniel Nicolás and Ramon Santamaria.
#
#   CONTRIBUTORS:
#       Ramon Santamaria:   Supervision, review, redesign, update and maintenance...
#       Kevin Gato:         Initial implementation of basic components (2014)
#       Daniel Nicolas:     Initial implementation of basic components (2014)
#       Albert Martos:      Review and testing of the library (2015)
#       Ian Eito:           Review and testing of the library (2015)
#       Sergio Martinez:    Review and testing of the library (2015)
#       Adria Arranz:       Testing and Implementation of additional controls (2018)
#       Jordi Jorba:        Testing and Implementation of additional controls (2018)
#
#
#   LICENSE: zlib/libpng
#
#   Copyright (c) 2014-2018 Ramon Santamaria (@raysan5)
#
#   This software is provided "as-is", without any express or implied warranty. In no event
#   will the authors be held liable for any damages arising from the use of this software.
#
#   Permission is granted to anyone to use this software for any purpose, including commercial
#   applications, and to alter it and redistribute it freely, subject to the following restrictions:
#
#     1. The origin of this software must not be misrepresented; you must not claim that you
#     wrote the original software. If you use this software in a product, an acknowledgment
#     in the product documentation would be appreciated but is not required.
#
#     2. Altered source versions must be plainly marked as such, and must not be misrepresented
#     as being the original software.
#
#     3. This notice may not be removed or altered from any source distribution.
#
# ---------------------------------------------------------------------------------------
# endregion

from typing import List
from enum import IntEnum
from . import (
    MOUSE_LEFT_BUTTON,
    Vector2,
    Vector3,
    Color,
    Rectangle,
    Texture2D,
    fade,
    check_collision_point_rec,
    format_text,
    measure_text,
    draw_text,
    draw_rectangle,
    draw_rectangle_rec,
    get_mouse_position,
    is_mouse_button_down,
    is_mouse_button_released,
    is_mouse_button_up,
    get_key_pressed,
    is_key_pressed,
    is_key_down,
    draw_rectangle_lines,
    draw_rectangle_lines_ex,
    draw_rectangle_gradient_h,
    draw_rectangle_gradient_v,
    draw_rectangle_gradient_ex,
    draw_texture,
    load_image,
    get_image_data,
    unload_image
)

__all__ = [

]

class GuiProperty(IntEnum):
    # -------------------------------------------
    # NOTE: This first set of properties is for general style,
    # following control-specific properties overwritte those styles
    DEFAULT_BACKGROUND_COLOR = 0
    DEFAULT_LINES_COLOR = 1
    DEFAULT_TEXT_FONT = 2
    DEFAULT_TEXT_SIZE = 3
    DEFAULT_BORDER_WIDTH = 4
    DEFAULT_BORDER_COLOR_NORMAL = 5
    DEFAULT_BASE_COLOR_NORMAL = 6
    DEFAULT_TEXT_COLOR_NORMAL = 7
    DEFAULT_BORDER_COLOR_FOCUSED = 8
    DEFAULT_BASE_COLOR_FOCUSED = 9
    DEFAULT_TEXT_COLOR_FOCUSED = 10
    DEFAULT_BORDER_COLOR_PRESSED = 11
    DEFAULT_BASE_COLOR_PRESSED = 12
    DEFAULT_TEXT_COLOR_PRESSED = 13
    DEFAULT_BORDER_COLOR_DISABLED = 14
    DEFAULT_BASE_COLOR_DISABLED = 15
    DEFAULT_TEXT_COLOR_DISABLED = 16
    # -------------------------------------------
    # Label
    LABEL_TEXT_COLOR_NORMAL = 17
    LABEL_TEXT_COLOR_FOCUSED = 18
    LABEL_TEXT_COLOR_PRESSED = 19
    LABEL_TEXT_COLOR_DISABLED = 20
    # Button
    BUTTON_BORDER_WIDTH = 21
    BUTTON_BORDER_COLOR_NORMAL = 22
    BUTTON_BASE_COLOR_NORMAL = 23
    BUTTON_TEXT_COLOR_NORMAL = 24
    BUTTON_BORDER_COLOR_FOCUSED = 25
    BUTTON_BASE_COLOR_FOCUSED = 26
    BUTTON_TEXT_COLOR_FOCUSED = 27
    BUTTON_BORDER_COLOR_PRESSED = 28
    BUTTON_BASE_COLOR_PRESSED = 29
    BUTTON_TEXT_COLOR_PRESSED = 30
    BUTTON_BORDER_COLOR_DISABLED = 31
    BUTTON_BASE_COLOR_DISABLED = 32
    BUTTON_TEXT_COLOR_DISABLED = 33
    # Toggle
    TOGGLE_BORDER_WIDTH = 34
    TOGGLE_BORDER_COLOR_NORMAL = 35
    TOGGLE_BASE_COLOR_NORMAL = 36
    TOGGLE_TEXT_COLOR_NORMAL = 37
    TOGGLE_BORDER_COLOR_FOCUSED = 38
    TOGGLE_BASE_COLOR_FOCUSED = 39
    TOGGLE_TEXT_COLOR_FOCUSED = 40
    TOGGLE_BORDER_COLOR_PRESSED = 41
    TOGGLE_BASE_COLOR_PRESSED = 42
    TOGGLE_TEXT_COLOR_PRESSED = 43
    TOGGLE_BORDER_COLOR_DISABLED = 44
    TOGGLE_BASE_COLOR_DISABLED = 45
    TOGGLE_TEXT_COLOR_DISABLED = 46
    TOGGLEGROUP_PADDING = 47
    # Slider
    SLIDER_BORDER_WIDTH = 48
    SLIDER_SLIDER_WIDTH = 49
    SLIDER_BORDER_COLOR_NORMAL = 50
    SLIDER_BASE_COLOR_NORMAL = 51
    SLIDER_BORDER_COLOR_FOCUSED = 52
    SLIDER_BASE_COLOR_FOCUSED = 53
    SLIDER_BORDER_COLOR_PRESSED = 54
    SLIDER_BASE_COLOR_PRESSED = 55
    SLIDER_BORDER_COLOR_DISABLED = 56
    SLIDER_BASE_COLOR_DISABLED = 57
    # SliderBar
    SLIDERBAR_INNER_PADDING = 58
    SLIDERBAR_BORDER_WIDTH = 59
    SLIDERBAR_BORDER_COLOR_NORMAL = 60
    SLIDERBAR_BASE_COLOR_NORMAL = 61
    SLIDERBAR_BORDER_COLOR_FOCUSED = 62
    SLIDERBAR_BASE_COLOR_FOCUSED = 63
    SLIDERBAR_BORDER_COLOR_PRESSED = 64
    SLIDERBAR_BASE_COLOR_PRESSED = 65
    SLIDERBAR_BORDER_COLOR_DISABLED = 66
    SLIDERBAR_BASE_COLOR_DISABLED = 67
    # ProgressBar
    PROGRESSBAR_INNER_PADDING = 68
    PROGRESSBAR_BORDER_WIDTH = 69
    PROGRESSBAR_BORDER_COLOR_NORMAL = 70
    PROGRESSBAR_BASE_COLOR_NORMAL = 71
    PROGRESSBAR_BORDER_COLOR_FOCUSED = 72
    PROGRESSBAR_BASE_COLOR_FOCUSED = 73
    PROGRESSBAR_BORDER_COLOR_PRESSED = 74
    PROGRESSBAR_BASE_COLOR_PRESSED = 75
    PROGRESSBAR_BORDER_COLOR_DISABLED = 76
    PROGRESSBAR_BASE_COLOR_DISABLED = 77
    # ValueBox
    VALUEBOX_BUTTON_PADDING = 78
    VALUEBOX_BUTTONS_WIDTH = 79
    VALUEBOX_BORDER_COLOR_NORMAL = 80
    VALUEBOX_BASE_COLOR_NORMAL = 81
    VALUEBOX_TEXT_COLOR_NORMAL = 82
    VALUEBOX_BORDER_COLOR_FOCUSED = 83
    VALUEBOX_BASE_COLOR_FOCUSED = 84
    VALUEBOX_TEXT_COLOR_FOCUSED = 85
    VALUEBOX_BORDER_COLOR_PRESSED = 86
    VALUEBOX_BASE_COLOR_PRESSED = 87
    VALUEBOX_TEXT_COLOR_PRESSED = 88
    VALUEBOX_BORDER_COLOR_DISABLED = 89
    VALUEBOX_BASE_COLOR_DISABLED = 90
    VALUEBOX_TEXT_COLOR_DISABLED = 91
    # ComboBox
    COMBOBOX_BORDER_WIDTH = 92
    COMBOBOX_BUTTON_PADDING = 93
    COMBOBOX_SELECTOR_WIDTH = 94
    COMBOBOX_BORDER_COLOR_NORMAL = 95
    COMBOBOX_BASE_COLOR_NORMAL = 96
    COMBOBOX_TEXT_COLOR_NORMAL = 97
    COMBOBOX_BORDER_COLOR_FOCUSED = 98
    COMBOBOX_BASE_COLOR_FOCUSED = 99
    COMBOBOX_TEXT_COLOR_FOCUSED = 100
    COMBOBOX_BORDER_COLOR_PRESSED = 101
    COMBOBOX_BASE_COLOR_PRESSED = 102
    COMBOBOX_TEXT_COLOR_PRESSED = 103
    COMBOBOX_BORDER_COLOR_DISABLED = 104
    COMBOBOX_BASE_COLOR_DISABLED = 105
    COMBOBOX_TEXT_COLOR_DISABLED = 106
    # CheckBox
    CHECKBOX_BORDER_WIDTH = 107
    CHECKBOX_INNER_PADDING = 108
    CHECKBOX_BORDER_COLOR_NORMAL = 109
    CHECKBOX_BASE_COLOR_NORMAL = 110
    CHECKBOX_BORDER_COLOR_FOCUSED = 111
    CHECKBOX_BASE_COLOR_FOCUSED = 112
    CHECKBOX_BORDER_COLOR_PRESSED = 113
    CHECKBOX_BASE_COLOR_PRESSED = 114
    CHECKBOX_BORDER_COLOR_DISABLED = 115
    CHECKBOX_BASE_COLOR_DISABLED = 116
    # TextBox
    TEXTBOX_BORDER_WIDTH = 117
    TEXTBOX_BORDER_COLOR_NORMAL = 118
    TEXTBOX_BASE_COLOR_NORMAL = 119
    TEXTBOX_TEXT_COLOR_NORMAL = 120
    TEXTBOX_BORDER_COLOR_FOCUSED = 121
    TEXTBOX_BASE_COLOR_FOCUSED = 122
    TEXTBOX_TEXT_COLOR_FOCUSED = 123
    TEXTBOX_BORDER_COLOR_PRESSED = 124
    TEXTBOX_BASE_COLOR_PRESSED = 125
    TEXTBOX_TEXT_COLOR_PRESSED = 126
    TEXTBOX_BORDER_COLOR_DISABLED = 127
    TEXTBOX_BASE_COLOR_DISABLED = 128
    TEXTBOX_TEXT_COLOR_DISABLED = 129
    # ColorPicker
    COLORPICKER_BARS_THICK = 130
    COLORPICKER_BARS_PADDING = 131
    COLORPICKER_BORDER_COLOR_NORMAL = 132
    COLORPICKER_BASE_COLOR_NORMAL = 133
    COLORPICKER_BORDER_COLOR_FOCUSED = 134
    COLORPICKER_BASE_COLOR_FOCUSED = 135
    COLORPICKER_BORDER_COLOR_PRESSED = 136
    COLORPICKER_BASE_COLOR_PRESSED = 137
    COLORPICKER_BORDER_COLOR_DISABLED = 138
    COLORPICKER_BASE_COLOR_DISABLED = 139
    # ListView
    LISTVIEW_ELEMENTS_HEIGHT = 140
    LISTVIEW_ELEMENTS_PADDING = 141
    LISTVIEW_BAR_WIDTH = 142
    LISTVIEW_BORDER_COLOR_NORMAL = 143
    LISTVIEW_BASE_COLOR_NORMAL = 144
    LISTVIEW_TEXT_COLOR_NORMAL = 145
    LISTVIEW_BORDER_COLOR_FOCUSED = 146
    LISTVIEW_BASE_COLOR_FOCUSED = 147
    LISTVIEW_TEXT_COLOR_FOCUSED = 148
    LISTVIEW_BORDER_COLOR_PRESSED = 149
    LISTVIEW_BASE_COLOR_PRESSED = 150
    LISTVIEW_TEXT_COLOR_PRESSED = 151
    LISTVIEW_BORDER_COLOR_DISABLED = 152
    LISTVIEW_BASE_COLOR_DISABLED = 153
    LISTVIEW_TEXT_COLOR_DISABLED = 154

DEFAULT_BACKGROUND_COLOR = GuiProperty.DEFAULT_BACKGROUND_COLOR
DEFAULT_LINES_COLOR = GuiProperty.DEFAULT_LINES_COLOR
DEFAULT_TEXT_FONT = GuiProperty.DEFAULT_TEXT_FONT
DEFAULT_TEXT_SIZE = GuiProperty.DEFAULT_TEXT_SIZE
DEFAULT_BORDER_WIDTH = GuiProperty.DEFAULT_BORDER_WIDTH
DEFAULT_BORDER_COLOR_NORMAL = GuiProperty.DEFAULT_BORDER_COLOR_NORMAL
DEFAULT_BASE_COLOR_NORMAL = GuiProperty.DEFAULT_BASE_COLOR_NORMAL
DEFAULT_TEXT_COLOR_NORMAL = GuiProperty.DEFAULT_TEXT_COLOR_NORMAL
DEFAULT_BORDER_COLOR_FOCUSED = GuiProperty.DEFAULT_BORDER_COLOR_FOCUSED
DEFAULT_BASE_COLOR_FOCUSED = GuiProperty.DEFAULT_BASE_COLOR_FOCUSED
DEFAULT_TEXT_COLOR_FOCUSED = GuiProperty.DEFAULT_TEXT_COLOR_FOCUSED
DEFAULT_BORDER_COLOR_PRESSED = GuiProperty.DEFAULT_BORDER_COLOR_PRESSED
DEFAULT_BASE_COLOR_PRESSED = GuiProperty.DEFAULT_BASE_COLOR_PRESSED
DEFAULT_TEXT_COLOR_PRESSED = GuiProperty.DEFAULT_TEXT_COLOR_PRESSED
DEFAULT_BORDER_COLOR_DISABLED = GuiProperty.DEFAULT_BORDER_COLOR_DISABLED
DEFAULT_BASE_COLOR_DISABLED = GuiProperty.DEFAULT_BASE_COLOR_DISABLED
DEFAULT_TEXT_COLOR_DISABLED = GuiProperty.DEFAULT_TEXT_COLOR_DISABLED
# -------------------------------------------
# Label
LABEL_TEXT_COLOR_NORMAL = GuiProperty.LABEL_TEXT_COLOR_NORMAL
LABEL_TEXT_COLOR_FOCUSED = GuiProperty.LABEL_TEXT_COLOR_FOCUSED
LABEL_TEXT_COLOR_PRESSED = GuiProperty.LABEL_TEXT_COLOR_PRESSED
LABEL_TEXT_COLOR_DISABLED = GuiProperty.LABEL_TEXT_COLOR_DISABLED
# Button
BUTTON_BORDER_WIDTH = GuiProperty.BUTTON_BORDER_WIDTH
BUTTON_BORDER_COLOR_NORMAL = GuiProperty.BUTTON_BORDER_COLOR_NORMAL
BUTTON_BASE_COLOR_NORMAL = GuiProperty.BUTTON_BASE_COLOR_NORMAL
BUTTON_TEXT_COLOR_NORMAL = GuiProperty.BUTTON_TEXT_COLOR_NORMAL
BUTTON_BORDER_COLOR_FOCUSED = GuiProperty.BUTTON_BORDER_COLOR_FOCUSED
BUTTON_BASE_COLOR_FOCUSED = GuiProperty.BUTTON_BASE_COLOR_FOCUSED
BUTTON_TEXT_COLOR_FOCUSED = GuiProperty.BUTTON_TEXT_COLOR_FOCUSED
BUTTON_BORDER_COLOR_PRESSED = GuiProperty.BUTTON_BORDER_COLOR_PRESSED
BUTTON_BASE_COLOR_PRESSED = GuiProperty.BUTTON_BASE_COLOR_PRESSED
BUTTON_TEXT_COLOR_PRESSED = GuiProperty.BUTTON_TEXT_COLOR_PRESSED
BUTTON_BORDER_COLOR_DISABLED = GuiProperty.BUTTON_BORDER_COLOR_DISABLED
BUTTON_BASE_COLOR_DISABLED = GuiProperty.BUTTON_BASE_COLOR_DISABLED
BUTTON_TEXT_COLOR_DISABLED = GuiProperty.BUTTON_TEXT_COLOR_DISABLED
# Toggle
TOGGLE_BORDER_WIDTH = GuiProperty.TOGGLE_BORDER_WIDTH
TOGGLE_BORDER_COLOR_NORMAL = GuiProperty.TOGGLE_BORDER_COLOR_NORMAL
TOGGLE_BASE_COLOR_NORMAL = GuiProperty.TOGGLE_BASE_COLOR_NORMAL
TOGGLE_TEXT_COLOR_NORMAL = GuiProperty.TOGGLE_TEXT_COLOR_NORMAL
TOGGLE_BORDER_COLOR_FOCUSED = GuiProperty.TOGGLE_BORDER_COLOR_FOCUSED
TOGGLE_BASE_COLOR_FOCUSED = GuiProperty.TOGGLE_BASE_COLOR_FOCUSED
TOGGLE_TEXT_COLOR_FOCUSED = GuiProperty.TOGGLE_TEXT_COLOR_FOCUSED
TOGGLE_BORDER_COLOR_PRESSED = GuiProperty.TOGGLE_BORDER_COLOR_PRESSED
TOGGLE_BASE_COLOR_PRESSED = GuiProperty.TOGGLE_BASE_COLOR_PRESSED
TOGGLE_TEXT_COLOR_PRESSED = GuiProperty.TOGGLE_TEXT_COLOR_PRESSED
TOGGLE_BORDER_COLOR_DISABLED = GuiProperty.TOGGLE_BORDER_COLOR_DISABLED
TOGGLE_BASE_COLOR_DISABLED = GuiProperty.TOGGLE_BASE_COLOR_DISABLED
TOGGLE_TEXT_COLOR_DISABLED = GuiProperty.TOGGLE_TEXT_COLOR_DISABLED
TOGGLEGROUP_PADDING = GuiProperty.TOGGLEGROUP_PADDING
# Slider
SLIDER_BORDER_WIDTH = GuiProperty.SLIDER_BORDER_WIDTH
SLIDER_SLIDER_WIDTH = GuiProperty.SLIDER_SLIDER_WIDTH
SLIDER_BORDER_COLOR_NORMAL = GuiProperty.SLIDER_BORDER_COLOR_NORMAL
SLIDER_BASE_COLOR_NORMAL = GuiProperty.SLIDER_BASE_COLOR_NORMAL
SLIDER_BORDER_COLOR_FOCUSED = GuiProperty.SLIDER_BORDER_COLOR_FOCUSED
SLIDER_BASE_COLOR_FOCUSED = GuiProperty.SLIDER_BASE_COLOR_FOCUSED
SLIDER_BORDER_COLOR_PRESSED = GuiProperty.SLIDER_BORDER_COLOR_PRESSED
SLIDER_BASE_COLOR_PRESSED = GuiProperty.SLIDER_BASE_COLOR_PRESSED
SLIDER_BORDER_COLOR_DISABLED = GuiProperty.SLIDER_BORDER_COLOR_DISABLED
SLIDER_BASE_COLOR_DISABLED = GuiProperty.SLIDER_BASE_COLOR_DISABLED
# SliderBar
SLIDERBAR_INNER_PADDING = GuiProperty.SLIDERBAR_INNER_PADDING
SLIDERBAR_BORDER_WIDTH = GuiProperty.SLIDERBAR_BORDER_WIDTH
SLIDERBAR_BORDER_COLOR_NORMAL = GuiProperty.SLIDERBAR_BORDER_COLOR_NORMAL
SLIDERBAR_BASE_COLOR_NORMAL = GuiProperty.SLIDERBAR_BASE_COLOR_NORMAL
SLIDERBAR_BORDER_COLOR_FOCUSED = GuiProperty.SLIDERBAR_BORDER_COLOR_FOCUSED
SLIDERBAR_BASE_COLOR_FOCUSED = GuiProperty.SLIDERBAR_BASE_COLOR_FOCUSED
SLIDERBAR_BORDER_COLOR_PRESSED = GuiProperty.SLIDERBAR_BORDER_COLOR_PRESSED
SLIDERBAR_BASE_COLOR_PRESSED = GuiProperty.SLIDERBAR_BASE_COLOR_PRESSED
SLIDERBAR_BORDER_COLOR_DISABLED = GuiProperty.SLIDERBAR_BORDER_COLOR_DISABLED
SLIDERBAR_BASE_COLOR_DISABLED = GuiProperty.SLIDERBAR_BASE_COLOR_DISABLED
# ProgressBar
PROGRESSBAR_INNER_PADDING = GuiProperty.PROGRESSBAR_INNER_PADDING
PROGRESSBAR_BORDER_WIDTH = GuiProperty.PROGRESSBAR_BORDER_WIDTH
PROGRESSBAR_BORDER_COLOR_NORMAL = GuiProperty.PROGRESSBAR_BORDER_COLOR_NORMAL
PROGRESSBAR_BASE_COLOR_NORMAL = GuiProperty.PROGRESSBAR_BASE_COLOR_NORMAL
PROGRESSBAR_BORDER_COLOR_FOCUSED = GuiProperty.PROGRESSBAR_BORDER_COLOR_FOCUSED
PROGRESSBAR_BASE_COLOR_FOCUSED = GuiProperty.PROGRESSBAR_BASE_COLOR_FOCUSED
PROGRESSBAR_BORDER_COLOR_PRESSED = GuiProperty.PROGRESSBAR_BORDER_COLOR_PRESSED
PROGRESSBAR_BASE_COLOR_PRESSED = GuiProperty.PROGRESSBAR_BASE_COLOR_PRESSED
PROGRESSBAR_BORDER_COLOR_DISABLED = GuiProperty.PROGRESSBAR_BORDER_COLOR_DISABLED
PROGRESSBAR_BASE_COLOR_DISABLED = GuiProperty.PROGRESSBAR_BASE_COLOR_DISABLED
# ValueBox
VALUEBOX_BUTTON_PADDING = GuiProperty.VALUEBOX_BUTTON_PADDING
VALUEBOX_BUTTONS_WIDTH = GuiProperty.VALUEBOX_BUTTONS_WIDTH
VALUEBOX_BORDER_COLOR_NORMAL = GuiProperty.VALUEBOX_BORDER_COLOR_NORMAL
VALUEBOX_BASE_COLOR_NORMAL = GuiProperty.VALUEBOX_BASE_COLOR_NORMAL
VALUEBOX_TEXT_COLOR_NORMAL = GuiProperty.VALUEBOX_TEXT_COLOR_NORMAL
VALUEBOX_BORDER_COLOR_FOCUSED = GuiProperty.VALUEBOX_BORDER_COLOR_FOCUSED
VALUEBOX_BASE_COLOR_FOCUSED = GuiProperty.VALUEBOX_BASE_COLOR_FOCUSED
VALUEBOX_TEXT_COLOR_FOCUSED = GuiProperty.VALUEBOX_TEXT_COLOR_FOCUSED
VALUEBOX_BORDER_COLOR_PRESSED = GuiProperty.VALUEBOX_BORDER_COLOR_PRESSED
VALUEBOX_BASE_COLOR_PRESSED = GuiProperty.VALUEBOX_BASE_COLOR_PRESSED
VALUEBOX_TEXT_COLOR_PRESSED = GuiProperty.VALUEBOX_TEXT_COLOR_PRESSED
VALUEBOX_BORDER_COLOR_DISABLED = GuiProperty.VALUEBOX_BORDER_COLOR_DISABLED
VALUEBOX_BASE_COLOR_DISABLED = GuiProperty.VALUEBOX_BASE_COLOR_DISABLED
VALUEBOX_TEXT_COLOR_DISABLED = GuiProperty.VALUEBOX_TEXT_COLOR_DISABLED
# ComboBox
COMBOBOX_BORDER_WIDTH = GuiProperty.COMBOBOX_BORDER_WIDTH
COMBOBOX_BUTTON_PADDING = GuiProperty.COMBOBOX_BUTTON_PADDING
COMBOBOX_SELECTOR_WIDTH = GuiProperty.COMBOBOX_SELECTOR_WIDTH
COMBOBOX_BORDER_COLOR_NORMAL = GuiProperty.COMBOBOX_BORDER_COLOR_NORMAL
COMBOBOX_BASE_COLOR_NORMAL = GuiProperty.COMBOBOX_BASE_COLOR_NORMAL
COMBOBOX_TEXT_COLOR_NORMAL = GuiProperty.COMBOBOX_TEXT_COLOR_NORMAL
COMBOBOX_BORDER_COLOR_FOCUSED = GuiProperty.COMBOBOX_BORDER_COLOR_FOCUSED
COMBOBOX_BASE_COLOR_FOCUSED = GuiProperty.COMBOBOX_BASE_COLOR_FOCUSED
COMBOBOX_TEXT_COLOR_FOCUSED = GuiProperty.COMBOBOX_TEXT_COLOR_FOCUSED
COMBOBOX_BORDER_COLOR_PRESSED = GuiProperty.COMBOBOX_BORDER_COLOR_PRESSED
COMBOBOX_BASE_COLOR_PRESSED = GuiProperty.COMBOBOX_BASE_COLOR_PRESSED
COMBOBOX_TEXT_COLOR_PRESSED = GuiProperty.COMBOBOX_TEXT_COLOR_PRESSED
COMBOBOX_BORDER_COLOR_DISABLED = GuiProperty.COMBOBOX_BORDER_COLOR_DISABLED
COMBOBOX_BASE_COLOR_DISABLED = GuiProperty.COMBOBOX_BASE_COLOR_DISABLED
COMBOBOX_TEXT_COLOR_DISABLED = GuiProperty.COMBOBOX_TEXT_COLOR_DISABLED
# CheckBox
CHECKBOX_BORDER_WIDTH = GuiProperty.CHECKBOX_BORDER_WIDTH
CHECKBOX_INNER_PADDING = GuiProperty.CHECKBOX_INNER_PADDING
CHECKBOX_BORDER_COLOR_NORMAL = GuiProperty.CHECKBOX_BORDER_COLOR_NORMAL
CHECKBOX_BASE_COLOR_NORMAL = GuiProperty.CHECKBOX_BASE_COLOR_NORMAL
CHECKBOX_BORDER_COLOR_FOCUSED = GuiProperty.CHECKBOX_BORDER_COLOR_FOCUSED
CHECKBOX_BASE_COLOR_FOCUSED = GuiProperty.CHECKBOX_BASE_COLOR_FOCUSED
CHECKBOX_BORDER_COLOR_PRESSED = GuiProperty.CHECKBOX_BORDER_COLOR_PRESSED
CHECKBOX_BASE_COLOR_PRESSED = GuiProperty.CHECKBOX_BASE_COLOR_PRESSED
CHECKBOX_BORDER_COLOR_DISABLED = GuiProperty.CHECKBOX_BORDER_COLOR_DISABLED
CHECKBOX_BASE_COLOR_DISABLED = GuiProperty.CHECKBOX_BASE_COLOR_DISABLED
# TextBox
TEXTBOX_BORDER_WIDTH = GuiProperty.TEXTBOX_BORDER_WIDTH
TEXTBOX_BORDER_COLOR_NORMAL = GuiProperty.TEXTBOX_BORDER_COLOR_NORMAL
TEXTBOX_BASE_COLOR_NORMAL = GuiProperty.TEXTBOX_BASE_COLOR_NORMAL
TEXTBOX_TEXT_COLOR_NORMAL = GuiProperty.TEXTBOX_TEXT_COLOR_NORMAL
TEXTBOX_BORDER_COLOR_FOCUSED = GuiProperty.TEXTBOX_BORDER_COLOR_FOCUSED
TEXTBOX_BASE_COLOR_FOCUSED = GuiProperty.TEXTBOX_BASE_COLOR_FOCUSED
TEXTBOX_TEXT_COLOR_FOCUSED = GuiProperty.TEXTBOX_TEXT_COLOR_FOCUSED
TEXTBOX_BORDER_COLOR_PRESSED = GuiProperty.TEXTBOX_BORDER_COLOR_PRESSED
TEXTBOX_BASE_COLOR_PRESSED = GuiProperty.TEXTBOX_BASE_COLOR_PRESSED
TEXTBOX_TEXT_COLOR_PRESSED = GuiProperty.TEXTBOX_TEXT_COLOR_PRESSED
TEXTBOX_BORDER_COLOR_DISABLED = GuiProperty.TEXTBOX_BORDER_COLOR_DISABLED
TEXTBOX_BASE_COLOR_DISABLED = GuiProperty.TEXTBOX_BASE_COLOR_DISABLED
TEXTBOX_TEXT_COLOR_DISABLED = GuiProperty.TEXTBOX_TEXT_COLOR_DISABLED
# ColorPicker
COLORPICKER_BARS_THICK = GuiProperty.COLORPICKER_BARS_THICK
COLORPICKER_BARS_PADDING = GuiProperty.COLORPICKER_BARS_PADDING
COLORPICKER_BORDER_COLOR_NORMAL = GuiProperty.COLORPICKER_BORDER_COLOR_NORMAL
COLORPICKER_BASE_COLOR_NORMAL = GuiProperty.COLORPICKER_BASE_COLOR_NORMAL
COLORPICKER_BORDER_COLOR_FOCUSED = GuiProperty.COLORPICKER_BORDER_COLOR_FOCUSED
COLORPICKER_BASE_COLOR_FOCUSED = GuiProperty.COLORPICKER_BASE_COLOR_FOCUSED
COLORPICKER_BORDER_COLOR_PRESSED = GuiProperty.COLORPICKER_BORDER_COLOR_PRESSED
COLORPICKER_BASE_COLOR_PRESSED = GuiProperty.COLORPICKER_BASE_COLOR_PRESSED
COLORPICKER_BORDER_COLOR_DISABLED = GuiProperty.COLORPICKER_BORDER_COLOR_DISABLED
COLORPICKER_BASE_COLOR_DISABLED = GuiProperty.COLORPICKER_BASE_COLOR_DISABLED
# ListView
LISTVIEW_ELEMENTS_HEIGHT = GuiProperty.LISTVIEW_ELEMENTS_HEIGHT
LISTVIEW_ELEMENTS_PADDING = GuiProperty.LISTVIEW_ELEMENTS_PADDING
LISTVIEW_BAR_WIDTH = GuiProperty.LISTVIEW_BAR_WIDTH
LISTVIEW_BORDER_COLOR_NORMAL = GuiProperty.LISTVIEW_BORDER_COLOR_NORMAL
LISTVIEW_BASE_COLOR_NORMAL = GuiProperty.LISTVIEW_BASE_COLOR_NORMAL
LISTVIEW_TEXT_COLOR_NORMAL = GuiProperty.LISTVIEW_TEXT_COLOR_NORMAL
LISTVIEW_BORDER_COLOR_FOCUSED = GuiProperty.LISTVIEW_BORDER_COLOR_FOCUSED
LISTVIEW_BASE_COLOR_FOCUSED = GuiProperty.LISTVIEW_BASE_COLOR_FOCUSED
LISTVIEW_TEXT_COLOR_FOCUSED = GuiProperty.LISTVIEW_TEXT_COLOR_FOCUSED
LISTVIEW_BORDER_COLOR_PRESSED = GuiProperty.LISTVIEW_BORDER_COLOR_PRESSED
LISTVIEW_BASE_COLOR_PRESSED = GuiProperty.LISTVIEW_BASE_COLOR_PRESSED
LISTVIEW_TEXT_COLOR_PRESSED = GuiProperty.LISTVIEW_TEXT_COLOR_PRESSED
LISTVIEW_BORDER_COLOR_DISABLED = GuiProperty.LISTVIEW_BORDER_COLOR_DISABLED
LISTVIEW_BASE_COLOR_DISABLED = GuiProperty.LISTVIEW_BASE_COLOR_DISABLED
LISTVIEW_TEXT_COLOR_DISABLED = GuiProperty.LISTVIEW_TEXT_COLOR_DISABLED


class GuiControlState(IntEnum):
    DISABLED = 0
    NORMAL = 1
    FOCUSED = 2
    PRESSED = 3

# Simulate C enums
DISABLED = GuiControlState.DISABLED
NORMAL = GuiControlState.NORMAL
FOCUSED = GuiControlState.FOCUSED
PRESSED = GuiControlState.PRESSED

WINDOW_PADDING = 2
WINDOW_BORDER_WIDTH = 1

GROUPBOX_LINE_THICK = 1
GROUPBOX_TEXT_PADDING = 10
GROUPBOX_PADDING = 2

NUM_PROPERTIES = 155

RAYGUI_STYLE_DEFAULT_LIGHT = [
    #----------------------------------------------
    0xf5f5f5ff,     # DEFAULT_BACKGROUND_COLOR
    0x90abb5ff,     # DEFAULT_LINES_COLOR
    0,              # DEFAULT_TEXT_FONT
    10,             # DEFAULT_TEXT_SIZE
    1,              # DEFAULT_BORDER_WIDTH
    0x838383ff,     # DEFAULT_BORDER_COLOR_NORMAL
    0xc9c9c9ff,     # DEFAULT_BASE_COLOR_NORMAL
    0x686868ff,     # DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # DEFAULT_TEXT_COLOR_DISABLED
    #----------------------------------------------
    0x686868ff,     # LABEL_TEXT_COLOR_NORMAL -----> DEFAULT_TEXT_COLOR_NORMAL
    0x6c9bbcff,     # LABEL_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x368bafff,     # LABEL_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xaeb7b8ff,     # LABEL_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x2,            # BUTTON_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x838383ff,     # BUTTON_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xc9c9c9ff,     # BUTTON_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x686868ff,     # BUTTON_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # BUTTON_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # BUTTON_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # BUTTON_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # BUTTON_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # BUTTON_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # BUTTON_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # BUTTON_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # BUTTON_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # BUTTON_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x1,            # TOGGLE_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x838383ff,     # TOGGLE_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xc9c9c9ff,     # TOGGLE_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x686868ff,     # TOGGLE_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # TOGGLE_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # TOGGLE_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # TOGGLE_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # TOGGLE_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # TOGGLE_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # TOGGLE_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # TOGGLE_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # TOGGLE_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # TOGGLE_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x2,            # TOGGLEGROUP_PADDING
    0x1,            # SLIDER_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x14,           # SLIDER_SLIDER_WIDTH 
    0x838383ff,     # SLIDER_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xc9c9c9ff,     # SLIDER_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x5bb2d9ff,     # SLIDER_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # SLIDER_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x0492c7ff,     # SLIDER_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # SLIDER_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0xb5c1c2ff,     # SLIDER_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # SLIDER_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1,            # SLIDERBAR_INNER_PADDING
    0x1,            # SLIDERBAR_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x838383ff,     # SLIDERBAR_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xc9c9c9ff,     # SLIDERBAR_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x5bb2d9ff,     # SLIDERBAR_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # SLIDERBAR_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x0492c7ff,     # SLIDERBAR_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # SLIDERBAR_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0xb5c1c2ff,     # SLIDERBAR_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # SLIDERBAR_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1,            # PROGRESSBAR_INNER_PADDING
    0x1,            # PROGRESSBAR_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x838383ff,     # PROGRESSBAR_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x97e8ffff,     # PROGRESSBAR_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_PRESSED
    0x5bb2d9ff,     # PROGRESSBAR_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # PROGRESSBAR_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x0492c7ff,     # PROGRESSBAR_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # PROGRESSBAR_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0xb5c1c2ff,     # PROGRESSBAR_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # PROGRESSBAR_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x2,            # VALUEBOX_BUTTON_PADDING
    0x23,           # VALUEBOX_BUTTONS_WIDTH
    0x838383ff,     # VALUEBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xf5f5f5ff,     # VALUEBOX_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x686868ff,     # VALUEBOX_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # VALUEBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # VALUEBOX_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # VALUEBOX_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # VALUEBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # VALUEBOX_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # VALUEBOX_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # VALUEBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # VALUEBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # VALUEBOX_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x1,            # COMBOBOX_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x2,            # COMBOBOX_BUTTON_PADDING
    0x23,           # COMBOBOX_SELECTOR_WIDTH
    0x838383ff,     # COMBOBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xc9c9c9ff,     # COMBOBOX_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x686868ff,     # COMBOBOX_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # COMBOBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # COMBOBOX_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # COMBOBOX_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # COMBOBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # COMBOBOX_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # COMBOBOX_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # COMBOBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # COMBOBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # COMBOBOX_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x1,            # CHECKBOX_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x1,            # CHECKBOX_INNER_PADDING
    0x838383ff,     # CHECKBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xf5f5f5ff,     # CHECKBOX_BASE_COLOR_NORMAL ----> DEFAULT_BACKGROUND_COLOR
    0x5bb2d9ff,     # CHECKBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x6c9bbcff,     # CHECKBOX_BASE_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # CHECKBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x368bafff,     # CHECKBOX_BASE_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # CHECKBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # CHECKBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1,            # TEXTBOX_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x838383ff,     # TEXTBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xf5f5f5ff,     # TEXTBOX_BASE_COLOR_NORMAL ----> DEFAULT_BACKGROUND_COLOR
    0x686868ff,     # TEXTBOX_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # TEXTBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # TEXTBOX_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # TEXTBOX_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # TEXTBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # TEXTBOX_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # TEXTBOX_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # TEXTBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # TEXTBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # TEXTBOX_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x14,           # COLORPICKER_BARS_THICK
    0xa,            # COLORPICKER_BARS_PADDING
    0x838383ff,     # COLORPICKER_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xf5f5f5ff,     # COLORPICKER_BASE_COLOR_NORMAL ----> DEFAULT_BACKGROUND_COLOR
    0x5bb2d9ff,     # COLORPICKER_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x6c9bbcff,     # COLORPICKER_BASE_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # COLORPICKER_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x368bafff,     # COLORPICKER_BASE_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # COLORPICKER_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # COLORPICKER_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1e,           # LISTVIEW_ELEMENTS_HEIGHT
    0x2,            # LISTVIEW_ELEMENTS_PADDING
    0xa,            # LISTVIEW_BAR_WIDTH
    0x838383ff,     # LISTVIEW_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0xf5f5f5ff,     # LISTVIEW_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x686868ff,     # LISTVIEW_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5bb2d9ff,     # LISTVIEW_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0xc9effeff,     # LISTVIEW_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6c9bbcff,     # LISTVIEW_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x0492c7ff,     # LISTVIEW_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97e8ffff,     # LISTVIEW_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x368bafff,     # LISTVIEW_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0xb5c1c2ff,     # LISTVIEW_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0xe6e9e9ff,     # LISTVIEW_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0xaeb7b8ff,     # LISTVIEW_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
]

RAYGUI_STYLE_DEFAULT_DARK = [
    #----------------------------------------------
    0x2b3a3aff,     # DEFAULT_BACKGROUND_COLOR
    0x638465ff,     # DEFAULT_LINES_COLOR
    0,              # DEFAULT_TEXT_FONT
    10,             # DEFAULT_TEXT_SIZE
    1,              # DEFAULT_BORDER_WIDTH
    0x60827dff,     # DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # DEFAULT_BASE_COLOR_NORMAL
    0x82a29fff,     # DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # DEFAULT_TEXT_COLOR_DISABLED
    #----------------------------------------------
    0x82a29fff,     # LABEL_TEXT_COLOR_NORMAL -----> DEFAULT_TEXT_COLOR_NORMAL
    0x6aa9b8ff,     # LABEL_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0x97af81ff,     # LABEL_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x666b69ff,     # LABEL_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x2,            # BUTTON_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x60827dff,     # BUTTON_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # BUTTON_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x82a29fff,     # BUTTON_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # BUTTON_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # BUTTON_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # BUTTON_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # BUTTON_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # BUTTON_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # BUTTON_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # BUTTON_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # BUTTON_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # BUTTON_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x1,            # TOGGLE_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x60827dff,     # TOGGLE_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # TOGGLE_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x82a29fff,     # TOGGLE_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # TOGGLE_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # TOGGLE_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # TOGGLE_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # TOGGLE_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # TOGGLE_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # TOGGLE_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # TOGGLE_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # TOGGLE_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # TOGGLE_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x2,            # TOGGLEGROUP_PADDING
    0x1,            # SLIDER_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x14,           # SLIDER_SLIDER_WIDTH 
    0x60827dff,     # SLIDER_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # SLIDER_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x5f9aa8ff,     # SLIDER_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # SLIDER_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0xa9cb8dff,     # SLIDER_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # SLIDER_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x5b6462ff,     # SLIDER_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # SLIDER_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1,            # SLIDERBAR_INNER_PADDING
    0x1,            # SLIDERBAR_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x60827dff,     # SLIDERBAR_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # SLIDERBAR_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x5f9aa8ff,     # SLIDERBAR_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # SLIDERBAR_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0xa9cb8dff,     # SLIDERBAR_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # SLIDERBAR_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x5b6462ff,     # SLIDERBAR_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # SLIDERBAR_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1,            # PROGRESSBAR_INNER_PADDING
    0x1,            # PROGRESSBAR_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x60827dff,     # PROGRESSBAR_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x3b6357ff,     # PROGRESSBAR_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_PRESSED
    0x5f9aa8ff,     # PROGRESSBAR_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # PROGRESSBAR_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0xa9cb8dff,     # PROGRESSBAR_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # PROGRESSBAR_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x5b6462ff,     # PROGRESSBAR_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # PROGRESSBAR_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x2,            # VALUEBOX_BUTTON_PADDING
    0x23,           # VALUEBOX_BUTTONS_WIDTH
    0x60827dff,     # VALUEBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # VALUEBOX_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x82a29fff,     # VALUEBOX_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # VALUEBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # VALUEBOX_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # VALUEBOX_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # VALUEBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # VALUEBOX_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # VALUEBOX_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # VALUEBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # VALUEBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # VALUEBOX_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x1,            # COMBOBOX_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x2,            # COMBOBOX_BUTTON_PADDING
    0x23,           # COMBOBOX_SELECTOR_WIDTH
    0x60827dff,     # COMBOBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2c3334ff,     # COMBOBOX_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x82a29fff,     # COMBOBOX_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # COMBOBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # COMBOBOX_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # COMBOBOX_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # COMBOBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # COMBOBOX_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # COMBOBOX_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # COMBOBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # COMBOBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # COMBOBOX_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x1,            # CHECKBOX_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x1,            # CHECKBOX_INNER_PADDING
    0x60827dff,     # CHECKBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2b3a3aff,     # CHECKBOX_BASE_COLOR_NORMAL ----> DEFAULT_BACKGROUND_COLOR
    0x5f9aa8ff,     # CHECKBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x6aa9b8ff,     # CHECKBOX_BASE_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # CHECKBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97af81ff,     # CHECKBOX_BASE_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # CHECKBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # CHECKBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1,            # TEXTBOX_BORDER_WIDTH ----> DEFAULT_BORDER_WIDTH
    0x60827dff,     # TEXTBOX_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2b3a3aff,     # TEXTBOX_BASE_COLOR_NORMAL ----> DEFAULT_BACKGROUND_COLOR
    0x82a29fff,     # TEXTBOX_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # TEXTBOX_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # TEXTBOX_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # TEXTBOX_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # TEXTBOX_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # TEXTBOX_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # TEXTBOX_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # TEXTBOX_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # TEXTBOX_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # TEXTBOX_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
    0x14,           # COLORPICKER_BARS_THICK
    0xa,            # COLORPICKER_BARS_PADDING
    0x60827dff,     # COLORPICKER_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2b3a3aff,     # COLORPICKER_BASE_COLOR_NORMAL ----> DEFAULT_BACKGROUND_COLOR
    0x5f9aa8ff,     # COLORPICKER_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x6aa9b8ff,     # COLORPICKER_BASE_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # COLORPICKER_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x97af81ff,     # COLORPICKER_BASE_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # COLORPICKER_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # COLORPICKER_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x1e,           # LISTVIEW_ELEMENTS_HEIGHT
    0x2,            # LISTVIEW_ELEMENTS_PADDING
    0xa,            # LISTVIEW_BAR_WIDTH
    0x60827dff,     # LISTVIEW_BORDER_COLOR_NORMAL ----> DEFAULT_BORDER_COLOR_NORMAL
    0x2b3a3aff,     # LISTVIEW_BASE_COLOR_NORMAL ----> DEFAULT_BASE_COLOR_NORMAL
    0x82a29fff,     # LISTVIEW_TEXT_COLOR_NORMAL ----> DEFAULT_TEXT_COLOR_NORMAL
    0x5f9aa8ff,     # LISTVIEW_BORDER_COLOR_FOCUSED ----> DEFAULT_BORDER_COLOR_FOCUSED
    0x334e57ff,     # LISTVIEW_BASE_COLOR_FOCUSED ----> DEFAULT_BASE_COLOR_FOCUSED
    0x6aa9b8ff,     # LISTVIEW_TEXT_COLOR_FOCUSED ----> DEFAULT_TEXT_COLOR_FOCUSED
    0xa9cb8dff,     # LISTVIEW_BORDER_COLOR_PRESSED ----> DEFAULT_BORDER_COLOR_PRESSED
    0x3b6357ff,     # LISTVIEW_BASE_COLOR_PRESSED ----> DEFAULT_BASE_COLOR_PRESSED
    0x97af81ff,     # LISTVIEW_TEXT_COLOR_PRESSED ----> DEFAULT_TEXT_COLOR_PRESSED
    0x5b6462ff,     # LISTVIEW_BORDER_COLOR_DISABLED ----> DEFAULT_BORDER_COLOR_DISABLED
    0x2c3334ff,     # LISTVIEW_BASE_COLOR_DISABLED ----> DEFAULT_BASE_COLOR_DISABLED
    0x666b69ff,     # LISTVIEW_TEXT_COLOR_DISABLED ----> DEFAULT_TEXT_COLOR_DISABLED
]
# VALIGN_OFFSET macro
def valingn_offset(h: int) -> int:
    """Vertical alignment for pixel perfect."""
    return h % 2


def get_color(hexvalue: int) -> Color:
    pass


def color_to_int(color: Color) -> int:
    pass


def convert_hsv_to_rgb(hsv: Vector3) -> Vector3:
    pass


def convert_rgb_to_hsv(hsv: Vector3) -> Vector3:
    pass


class SingletonMeta(type):
    """
    Define an Instance operation that lets clients access its unique
    instance.
    """

    def __init__(cls, name, bases, attrs, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class Gui(metaclass=SingletonMeta):

    def __init__(self):
        self._state: GuiControlState = NORMAL
        self._alpha: float = 1.0
        self._style: list = RAYGUI_STYLE_DEFAULT_DARK

    def __getitem__(self, key: int):
        return self._style[key]

    def _button_logic(self, clicked: bool, point: Vector2, bounds: Rectangle) -> bool:
        # check button state
        if check_collision_point_rec(point, bounds):
            if is_mouse_button_down(MOUSE_LEFT_BUTTON):
                self._state = PRESSED
            elif is_mouse_button_released(MOUSE_LEFT_BUTTON):
                clicked = True
            else:
                self._state = FOCUSED
        return clicked

    def enable(self):
        """Enable GUI controls."""
        self._state = NORMAL

    def disable(self):
        """Disable GUI controls."""
        self._state = DISABLED
    
    def fade(self, alpha: float):
        """Set GUI controls transparency (ranges from 0.0 to 1.0)."""
        self._alpha = max(0.0, min(float(alpha), 1.0))

    @property
    def style(self) -> list:      # TODO: fix return type
        return self._style

    #region Container/Separator controls
    def window_box(self, bounds: Rectangle, text: str) -> bool:
        clicked: bool = False
        state: GuiControlState = self._state

        offset_x: int = 10
        text_width: int = measure_text(text, self[DEFAULT_TEXT_SIZE])
        # text_height: int = self[DEFAULT_TEXT_SIZE]

        if bounds.width < text_width + offset_x * 2 + 16:
            bounds.width = text_width + offset_x * 2 + 16
        statusbar: Rectangle = Rectangle(bounds.x, bounds.y, bounds.width, 24)
        if bounds.height < 50:
            bounds.height = 50

        # Update
        # ---------------------------------------
        if state is not DISABLED:
            mouse_point: Vector2 = get_mouse_position()

            # check button state
            clicked = self._button_logic(clicked, mouse_point, (statusbar.x + statusbar.width - 24, statusbar.y, 20, statusbar.height))

        # Draw
        # ---------------------------------------
        rec = (bounds.x + WINDOW_BORDER_WIDTH,
               bounds.y + WINDOW_BORDER_WIDTH,
               bounds.width - WINDOW_BORDER_WIDTH * 2,
               bounds.height - WINDOW_BORDER_WIDTH * 2)
        rec2 = (statusbar.x,
                statusbar.width - 27 + WINDOW_PADDING * 2,
                statusbar.y + 1 + WINDOW_PADDING,
                25 - WINDOW_PADDING * 3,
                statusbar.height - 2 - WINDOW_PADDING * 2)

        border_color, base_color, text_color = {
            NORMAL: ('default.border.color.normal', 'default.base.color.normal', 'default.text.color.normal'),
            FOCUSED: ('default.border.color.normal', 'default.base.color.focused', 'default.text.color.focused'),
            PRESSED: ('default.border.color.normal', 'default.base.color.pressed', 'default.text.color.pressed'),
            DISABLED: ('default.border.color.disabled', 'default.base.color.disabled', 'default.text.color.disabled'),
        }.get(self._state, ('default.border.color.normal', 'default.base.color.normal', 'default.text.color.normal'))

        draw_rectangle_lines_ex(bounds, WINDOW_BORDER_WIDTH, fade(get_color(self[border_color]), self._alpha))
        draw_rectangle_rec(rec, fade(get_color(self['default.background.color']), self._alpha))
        self.statusbar(statusbar, text, offset_x)
        draw_rectangle_rec(rec2, fade(get_color(self[base_color]), self._alpha))
        draw_text('x', statusbar.x + statusbar.width - 16, statusbar.y + statusbar.height / 2 - self['default.text.size'], self['default.text.size'], fade(get_color(self[text_color]), self._alpha))

        return clicked

    def group_box(self, bounds: Rectangle, text: str) -> None:
        """Group Box control with title name"""
        state: GuiControlState = self._state

        # Draw
        # ---------------------------------------
        if state is NORMAL:
            draw_rectangle(bounds.x, bounds.y, GROUPBOX_LINE_THICK, bounds.height, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
            draw_rectangle(bounds.x, bounds.y + bounds.height - 1, bounds.width, GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
            draw_rectangle(bounds.x + bounds.width - 1, bounds.y, GROUPBOX_LINE_THICK, bounds.height, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
            if text == '':
                draw_rectangle(bounds.x, bounds.y, bounds.width, GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
            else:
                draw_rectangle(bounds.x, bounds.y, GROUPBOX_TEXT_PADDING, GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
                draw_rectangle(bounds.x + 2 * GROUPBOX_TEXT_PADDING + measure_text(text, self[DEFAULT_TEXT_SIZE]), bounds.y, bounds.width - 2*GROUPBOX_TEXT_PADDING - measure_text(text, self[DEFAULT_TEXT_SIZE]), GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
                draw_text(text, bounds.x + GROUPBOX_TEXT_PADDING + 2 * GROUPBOX_PADDING, bounds.y - 2*GROUPBOX_PADDING - GROUPBOX_LINE_THICK, self[DEFAULT_TEXT_SIZE], fade(get_color(self[LABEL_TEXT_COLOR_NORMAL]), self._alpha))

        elif state is DISABLED:
            draw_rectangle(bounds.x, bounds.y, GROUPBOX_LINE_THICK, bounds.height, fade(get_color(self[DEFAULT_BORDER_COLOR_DISABLED]), self._alpha))
            draw_rectangle(bounds.x, bounds.y + bounds.height - 1, bounds.width, GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_BORDER_COLOR_DISABLED]), self._alpha))
            draw_rectangle(bounds.x + bounds.width - 1, bounds.y, GROUPBOX_LINE_THICK, bounds.height, fade(get_color(self[DEFAULT_BORDER_COLOR_DISABLED]), self._alpha))
            if text == '':
                draw_rectangle(bounds.x, bounds.y, bounds.width, GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
            else:
                draw_rectangle(bounds.x, bounds.y, GROUPBOX_TEXT_PADDING, GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_BORDER_COLOR_DISABLED]), self._alpha))
                draw_rectangle(bounds.x + 2 * GROUPBOX_TEXT_PADDING + measure_text(text, self[DEFAULT_TEXT_SIZE]), bounds.y, bounds.width - 2 * GROUPBOX_TEXT_PADDING - measure_text(text, self[DEFAULT_TEXT_SIZE]), GROUPBOX_LINE_THICK, fade(get_color(self[DEFAULT_BORDER_COLOR_DISABLED]), self._alpha))
                draw_text(text, bounds.x + GROUPBOX_TEXT_PADDING + 2 * GROUPBOX_PADDING, bounds.y - 2 * GROUPBOX_PADDING - GROUPBOX_LINE_THICK, self[DEFAULT_TEXT_SIZE], fade(get_color(self[LABEL_TEXT_COLOR_DISABLED]), self._alpha))

    def line(self, bounds: Rectangle, thick: int):
        state: GuiControlState = self._state
        line: Rectangle = Rectangle(bounds.x, bounds.y + bounds.height / 2 - thick / 2, bounds.width, thick)
        
        if state in (NORMAL, FOCUSED, PRESSED):
            draw_rectangle_lines_ex(line, thick, fade(get_color(self[DEFAULT_LINES_COLOR]), self._alpha))
        elif state is DISABLED:
            draw_rectangle_lines_ex(line, thick, fade(get_color(self[DEFAULT_BORDER_COLOR_DISABLED]), self._alpha))

    def panel(self, bounds: Rectangle) -> None:
        pass
    
    def scroll_panel(self, bounds: Rectangle, content: Rectangle, view_scroll: Vector2) -> Vector2:
        pass
    #endregion

    #region Basic controls set'
    def label(self, bounds: Rectangle, text: str) -> None:
        pass

    def button(self, bounds: Rectangle, text: str) -> bool:
        pass

    def label_button(self, bounds: Rectangle, text: str) -> bool:
        pass

    def image_button(self, bounds: Rectangle, texture: Texture2D) -> bool:
        pass

    def image_button_ex(self, bounds: Rectangle, texture: Texture2D, text_source: Rectangle, text: str) -> bool:
        pass

    def toggle_button(self, bounds: Rectangle, text: str, toggle: bool) -> bool:
        pass

    def toggle_group(self, bounds: Rectangle, text: str, count: int, active: int) -> int:
        pass

    def checkbox(self, bounds: Rectangle, checked: bool) -> bool:
        pass

    def checkbox_ex(self, bounds: Rectangle, checked: bool, text: str) -> bool:
        pass

    def combobox(self, bounds: Rectangle, text: str, count: int, active: int) -> int:
        pass

    def dropdown_box(self, bounds: Rectangle, text: str, count: int, active: int) -> int:
        pass

    def spinner(self, bounds: Rectangle, value: int, max_value: int, btn_width: int) -> int:
        pass

    def valuebox(self, bounds: Rectangle, value: int, max_value: int) -> int:
        pass

    def textbox(self, bounds: Rectangle, text: str, text_size: int, free_edit: bool) -> bool:
        pass

    def textbox_multi(self, bounds: Rectangle, text: str, text_size: int, edit_mode: bool) -> bool:
        pass

    def slider(self, bounds: Rectangle, value: float, min_value: float, max_value: float) -> float:
        pass

    def slider_ex(self, bounds: Rectangle, value: float, min_value: float, max_value: float, text: str, show_value: bool) -> float:
        pass

    def sliderbar(self, bounds: Rectangle, value: float, min_value: float, max_value: float) -> float:
        pass

    def sliderbar_ex(self, bounds: Rectangle, value: float, min_value: float, max_value: float, text: str, show_value: bool) -> float:
        pass

    def progressbar(self, bounds: Rectangle, value: float, min_value: float, max_value: float) -> float:
        pass

    def progressbar_ex(self, bounds: Rectangle, value: float, min_value: float, max_value: float, show_value: bool) -> float:
        pass

    def statusbar(self, bounds: Rectangle, text: str, offset_x: int) -> None:
        pass

    def dummy_rec(self, bounds: Rectangle, text: str) -> None:
        pass
    #endregion

    #region advance controls set
    def list_element(self, bounds: Rectangle, text: str, active: bool) -> bool:
        pass

    def listview(self, bounds: Rectangle, text: str, count: int, active: int) -> int:
        pass

    def color_picker(self, bounds: Rectangle, color: Color) -> Color:
        pass

    def message_box(self, bounds: Rectangle, window_title: str, message: str) -> bool:
        pass

    def load_style(self, file_name: str, binary: bool) -> None:
        pass

    def save_style(self, file_name: str) -> None:
        pass

    def load_style_image(self, file_name: str) -> None:
        pass

    def load_style_palette(self, palette: List[Color]) -> None:
        pass

    def update_style_complete(self) -> None:
        pass
    #endregion
